class Node:
    """
    The node class represent a node in TSP problem. It is implemented as a doubly linked list, where each node has its predecessor and successor node defined.
    """

    def __init__(self):
        """
        Initialize a node
        """
        # the id of the node (this unique integer will help to distinguish between repeated coordinates nodes and also to build the distance matrix). It will be updated when initializing a tour
        self.id = -1

        # the position (index) of the node in a tour (to be update when initializing the tour)
        # the position can be either a positive or negative number, but is must be incremented by 1
        self.pos = -1

        # the predecessor and successor nodes at a tour (to be update when initializing the tour)
        self.pred = None
        self.succ = None

    def __eq__(self, other):
        """
        Equal comparison method between two nodes, which will be true when node ids are equal.

        :param other: other node of comparison
        :type other: Node
        :return: a boolean indicating if both nodes are equal
        :rtype: boolean
        """
        # check if other node is not None
        if (other):
            return (self.id == other.id)
        else:
            return False

    def __gt__(self, other):
        """
        Greater than comparison method between two nodes. This is relevant when building edges, to always keep same node ordering (for simmetric TSP).

        :param other: other node of comparison
        :type other: Node
        :return: a boolean indicating if node is greater than other node
        :rtype: boolean
        """
        return (self.id > other.id)

    def __hash__(self):
        """
        Hashing Node object is required to allow comparison of Edge, which are elements made of Nodes. The hash is performed in the id value.
        """
        return hash(self.id)

    def __str__(self):
        """
        The display string when printing the object

        :return: the display string
        :rtype: str
        """
        return f"({self.id})"

    def __repr__(self):
        """
        The display string when printing the object

        :return: the display string
        :rtype: str
        """
        return f"({self.id})"


class Node2D(Node):
    """
    The node 2D class represent a node in 2D cartesian space.
    """

    def __init__(self, x, y):
        """
        Initialize a node with its cartesian values

        :param x: the "x" coordinate value
        :type x: float
        :param y: the "y" coordinate value
        :type y: float
        """

        # initialize the super class
        super().__init__()

        # initialize the node values
        self.x = x
        self.y = y

    def __str__(self):
        """
        The display string when printing the object

        :return: the display string
        :rtype: str
        """
        return f"{self.id}:({self.x},{self.y})"

    def __repr__(self):
        """
        The display string when printing the object

        :return: the display string
        :rtype: str
        """
        return f"{self.id}:({self.x},{self.y})"


class Node3D(Node):
    """
    The node 3D class represent a node in 3D cartesian space.
    """

    def __init__(self, x, y, z):
        """
        Initialize a node with its cartesian values

        :param x: the "x" coordinate value
        :type x: float
        :param y: the "y" coordinate value
        :type y: float
        :param z: the "z" coordinate value
        :type z: float
        """

        # initialize the super class
        super().__init__()

        # initialize the node values
        self.x = x
        self.y = y
        self.z = z

    def __str__(self):
        """
        The display string when printing the object

        :return: the display string
        :rtype: str
        """
        return f"{self.id}:({self.x},{self.y},{self.z})"

    def __repr__(self):
        """
        The display string when printing the object

        :return: the display string
        :rtype: str
        """
        return f"{self.id}:({self.x},{self.y},{self.z})"


class NodePivot(Node):
    """
    The pivot node is a dummy node used at hamiltonian path tours, where edges containing these nodes will have zero cost
    """
    pass



class Edge:
    """
    The edge class represent an edge in space. The edge is a connection between two nodes. The connection is simmetric, i.e, the Edge(AB) == Edge(BA). This simmetric property is implemented in initialization method.
    """

    def __init__(self, n1, n2):
        """
        Initialize an edge with its nodes. The node order (n1/n2) is switched based on node values to guarantee the simmetric property of TSP edges, i.e, {n1,n2} == {n2,n1}. By comparing the nodes, the order will always be the same, so is not necessary to check node ordering later.

        :param n1: the first node
        :type n1: Node
        :param n2: the second node
        :type n2: Node
        """

        # asserting that a valid edge can't connect to same node
        assert(n1 != n2)

        if n1 < n2:
            self.n1 = n1
            self.n2 = n2
        # switch the ordering
        else:
            self.n1 = n2
            self.n2 = n1

    def __eq__(self, other):
        """
        Comparison method between two edges. The edge is equal if its nodes are equal. The edge {n1,n2} will be equal to {n2,n1}. It is required to check the inverted conditions since node comparison will also compare node id (to be able to create two equal coordinates nodes - relevant for a280.tsp) and in this situation, node comparison would be false.

        :param other: the other edge to compare
        :type other: Edge
        :return: a boolean indicating if both edges are the same
        :rtype: boolean
        """
        return (self.n1 == other.n1 and self.n2 == other.n2) or (self.n1 == other.n2 and self.n2 == other.n1)

    def __hash__(self):
        """
        Hashing Edge object is required to allow building sets of edges, which are "fast" objects used in LK joined and broken edges sets. The hash is performed in the tuple of node 1 and node 2 id values.
        """
        return hash((self.n1.id, self.n2.id))

    def __str__(self):
        """
        The display string when printing the object

        :return: the display string
        :rtype: str
        """
        return f"({self.n1},{self.n2})"

    def __repr__(self):
        """
        The display string when printing the object

        :return: the display string
        :rtype: str
        """
        return f"({self.n1},{self.n2})"



import logging
from lk_heuristic.models.node import Node2D, Node3D

# create a logger object
logger = logging.getLogger(__name__)


def import_tsp_file(tsp_file):
    """
    Import a .tsp file containing tsp definition and parse tsp header to a dictionary and tsp nodes to a list. Returns both objects in a tuple.

    *This function only works for .tsp instances of "TYPE" = TSP (simmetric tsp) and for "EDGE_WEIGH_TYPE" = EUC_2D or EUC_3D, which are the tsp problems that solver can handle currently.

    :param tsp_file: the .tsp file path
    :type tsp_file: str
    :return: a tuple with the header dict definition and a list of nodes
    :rtype: dict
    """

    # set the allowed type and edge weight for tsp
    allowed_types = ["TSP"]
    allowed_edge_weights = ["EUC_2D", "EUC_3D"]

    # return objects
    tsp_header = {}  # the .tsp header dict to be returned by the function
    nodes = []  # the list of nodes to be parsed from node coord section

    # a boolean value to be changed when node coord section is reached
    is_node_section = False

    # try to execute the import
    try:

        # open tsp file
        with open(tsp_file, "r+") as f:

            # read all tsp file lines
            tsp_lines = f.readlines()

            # loop through each tsp line
            for line in tsp_lines:

                # checking if coord section was already reached
                if is_node_section:

                    # check if problem is tsp (simmetric tsp) and edge_weight is either euc_2d or euc_3d
                    if tsp_header["TYPE"] in allowed_types and tsp_header["EDGE_WEIGHT_TYPE"] in allowed_edge_weights:

                        # check if line is not last end-of-file line
                        if line.strip() != "EOF":

                            # split the line using space separator
                            coords = line.replace("\n", "").split()

                            # create a node using node coord values and append it to list
                            if tsp_header["EDGE_WEIGHT_TYPE"] == "EUC_2D":
                                nodes.append(Node2D(float(coords[1]), float(coords[2])))
                            elif tsp_header["EDGE_WEIGHT_TYPE"] == "EUC_3D":
                                nodes.append(Node3D(float(coords[1]), float(coords[2]), float(coords[3])))

                # try to collect .tsp attributes
                else:

                    # add node coord section into dict
                    if line.strip() == "NODE_COORD_SECTION":
                        is_node_section = True

                    # split the line using ":" to get header attributes
                    line_split = line.split(":")

                    # check if split was made
                    if len(line_split) == 2:

                        # assign dict values
                        tsp_header[line_split[0].strip()] = line_split[1].strip()

    # print error to user
    except Exception as e:
        logger.error(f"Error during import of .tsp file: '{e}'")

    # return the .tsp dict and node values
    return tsp_header, nodes


def export_tsp_file(tsp_file_path, tsp_header, nodes, tour_type="cycle"):
    """
    Export a .tsp file using tsp_header header elements and list of nodes. Nodes are parsed into "NODES_COORD_SECTION" format.

    :param tsp_file_path: the name of tsp file to be exported
    :type tsp_file_path: str
    :param tsp_header: a dictionary containing header values for the .tsp file
    :type tsp_header: dict
    :param nodes: a list containing nodes to be parsed into tsp file
    :type nodes: list
    :param tour_type: the type of the tour (either 'path' or 'cycle')
    :type tour_type: str
    """

    # try to execute the export
    try:

        # create tsp file to write data
        with open(tsp_file_path, "w+") as f:

            # loop through each header value and write it to file
            for k, v in tsp_header.items():
                f.write(f"{k} : {v}\n")

            # write the node section
            f.write("NODE_COORD_SECTION\n")

            # removes pivot node when using hamiltonian path tour type
            # pivot node shall be the first node (from Tour get_nodes function)
            if tour_type == "path":
                nodes.pop(0)

            # loop through each node and write its values to file
            for i, node in enumerate(nodes):
                if type(node) == Node2D:
                    f.write(f"{i} {node.x} {node.y}\n")
                elif type(node) == Node3D:
                    f.write(f"{i} {node.x} {node.y} {node.z}\n")

            # write end of file line
            f.write("EOF\n")

    # print error to user
    except Exception as e:
        logger.error(f"Error during export of .tsp file: '{e}'")



import unittest
from lk_heuristic.models.node import Node2D, Node3D
from lk_heuristic.models.edge import Edge


class TestEdge(unittest.TestCase):
    """
    A class to perform testing on TSP edges
    """

    def setUp(self):
        """
        The method to be performed before every test. Used to setup some edges
        """

        # initialize 2D test nodes
        node_2d_1 = Node2D(0, 1)
        node_2d_1.id = 0
        node_2d_2 = Node2D(1, 0)
        node_2d_2.id = 1
        node_2d_3 = Node2D(0, 1)
        node_2d_3.id = 2

        #  a test edge
        self.edge_2d_1 = Edge(node_2d_1, node_2d_2)

        # an equal edge
        self.edge_2d_equal_1 = Edge(node_2d_1, node_2d_2)

        # the reversed edge
        self.edge_2d_reversed_1 = Edge(node_2d_2, node_2d_1)

        # edge with different node id
        self.edge_2d_almost_equal_1 = Edge(node_2d_1, node_2d_3)

        # initialize 3D test nodes
        node_3d_1 = Node3D(0, 1, 0)
        node_3d_1.id = 0
        node_3d_2 = Node3D(1, 0, 0)
        node_3d_2.id = 1
        node_3d_3 = Node3D(0, 1, 0)
        node_3d_3.id = 2

        #  a test edge
        self.edge_3d_1 = Edge(node_3d_1, node_3d_2)

        # an equal edge
        self.edge_3d_equal_1 = Edge(node_3d_1, node_3d_2)

        # the reversed edge
        self.edge_3d_reversed_1 = Edge(node_3d_2, node_3d_1)

        # edge with different node id
        self.edge_3d_almost_equal_1 = Edge(node_3d_1, node_3d_3)

    def test_edge_2d_equality(self):
        """
        Testing edge equality.
        """

        # equal edges
        self.assertEqual(self.edge_2d_1, self.edge_2d_equal_1)

        # edge (n1,n2) must be equal to edge (n2,n1) if ids are the same
        self.assertEqual(self.edge_2d_1, self.edge_2d_reversed_1)

        # edge (n1,n2) not equal to edge (n2,n1) if ids are not the same
        self.assertNotEqual(self.edge_2d_1, self.edge_2d_almost_equal_1)

    def test_edge_2d_valid(self):
        """
        Testing if edge is valid
        """

        # an edge isn't valid if both nodes are equal (it must connect different nodes)
        node = Node2D(1, 0)
        self.assertRaises(AssertionError, Edge, node, node)

    def test_edge_3d_equality(self):
        """
        Testing edge equality.
        """

        # equal edges
        self.assertEqual(self.edge_3d_1, self.edge_3d_equal_1)

        # edge (n1,n2) must be equal to edge (n2,n1) if ids are the same
        self.assertEqual(self.edge_3d_1, self.edge_3d_reversed_1)

        # edge (n1,n2) not equal to edge (n2,n1) if ids are not the same
        self.assertNotEqual(self.edge_3d_1, self.edge_3d_almost_equal_1)

    def test_edge_3d_valid(self):
        """
        Testing if edge is valid
        """

        # an edge isn't valid if both nodes are equal (it must connect different nodes)
        node = Node3D(1, 0, 0)
        self.assertRaises(AssertionError, Edge, node, node)



from random import shuffle
from lk_heuristic.models.node import NodePivot
from lk_heuristic.models.edge import Edge


class Tour:
    """
    The tour class represents a sequence of edges that starts at one node, visits all tour nodes and ends at same starting node (for 'cycle' tours) or not (for 'path' tours). The nodes will be a list of nodes in the ordering of visit, while edges is a set of edges (so ordering is not considered).
    """

    def __init__(self, nodes, t="cycle"):
        """
        A tour is made by a sequence of edges. Since edges are defined by sequence of nodes, the node sequence is used as input. The tour type ('t') can be either "cycle" (for the classic hamiltonian cycle tsp) or "path" (for the hamiltonian path tsp).

        :param nodes: the list of Nodes that makes the tour
        :type nodes: list
        :param t: the type of the tour
        :type t: str
        """

        # initialize the tour type
        self.t = t

        # set the tour nodes and initialize tour node parameters
        self.nodes = nodes
        self.set_nodes()

        # set tour edges using defined tour nodes
        self.edges = set()
        self.set_edges()

        # Initialize the cost value, to be update when calling the function set_cost using a distance matrix.
        self.cost = 0

        # set the size of the tour
        self.size = len(nodes)

        # init the swap stack
        # the swap stack is the "memory" of swap functions executed in a specific tour (relevant when required to undo the swaps). it is a tuple on the form (n1, n2, n3, n4, swap_operation)
        self.swap_stack = []

    def set_nodes(self):
        """
        Update tour nodes with specific related tour properties
        """

        # for hamiltonian path tsp, create the pivot node and append it to nodes list
        if self.t == "path":
            self.nodes.append(NodePivot())

        # loop through each node and set the tour properties
        for i in range(len(self.nodes)):

            # set node's successor node
            if (i == len(self.nodes) - 1):
                self.nodes[i].succ = self.nodes[0]  # the successor of the last node is the first node
            else:
                self.nodes[i].succ = self.nodes[i + 1]

            # set node's predecessor node
            self.nodes[i].pred = self.nodes[i - 1]

            # set node position
            self.nodes[i].pos = i

            # set node id
            self.nodes[i].id = i

    def set_edges(self):
        """
        Generate the tour edges using current tour nodes
        """

        # initialize the set of edges
        tour_edges = set()

        # initialize the first node
        curr_node = self.nodes[0]

        # loop through each node and add the edge of current node and its neighbor
        while(curr_node.succ != self.nodes[0]):
            tour_edges.add(Edge(curr_node, curr_node.succ))
            curr_node = curr_node.succ

        # add the closing edge
        tour_edges.add(Edge(curr_node, curr_node.succ))

        # populate the edges
        self.edges = tour_edges

    def set_cost(self, cost_matrix):
        """
        Set the tour cost by summing all tour edge costs from a cost matrix

        :param cost_matrix: the matrix with cost(i,j) between node id "i" and node id "j"
        :type node: list

        """
        # initialize the cost
        tour_cost = 0

        # initialize the first node
        curr_node = self.nodes[0]

        # loop through each node and add the edge of current node and its neighbor
        while(curr_node.succ != self.nodes[0]):
            tour_cost += cost_matrix[(curr_node.id, curr_node.succ.id)]
            curr_node = curr_node.succ

        # add the cost of closing the loop (last node -> first node)
        tour_cost += cost_matrix[(curr_node.id, curr_node.succ.id)]

        # assign the cost value
        self.cost = tour_cost

    def set_pos(self):
        """
        Set the 'pos' attribute of the nodes from a feasible tour. This function is relevant after performing unfeasible swaps that only reorder nodes pred/succ attribute but leaves the pos attribute incorrect. After converging to a feasible tour, call this function to update the pos attribute.
        """

        # initialization with first node
        curr_node = self.nodes[0]
        curr_node.pos = 0

        # loop until all nodes have been seen
        while curr_node.succ != self.nodes[0]:

            # update current node
            curr_node = curr_node.succ
            curr_node.pos = curr_node.pred.pos + 1

    def get_nodes(self):
        """
        Get the nodes inside the tour sorted by their sequence and return the list of nodes

        :return: the list of node indexes
        :rtype: list
        """

        # a set of nodes to check if all nodes were analyzed
        visited_nodes = set(self.nodes)

        # start a list of sorted nodes
        tour_nodes = []

        # get the start index for first node
        # when tour is hamiltonian path, the first node will be the pivot node (so when exporting the tsp file, this node is removed)
        start_idx = 0
        if self.t == "path":
            for i, node in enumerate(self.nodes):
                if type(node) == NodePivot:
                    start_idx = i
                    break

        # initialization with first node
        curr_node = self.nodes[start_idx]
        visited_nodes.remove(curr_node)

        # loop until all nodes have been seen (this is necessary for unfeasible tours, like two separated subtours)
        while len(visited_nodes) > 0:

            # loop through each tour node checking if it was not visited yet
            while curr_node.succ in visited_nodes:

                # append the node index to list of indexes
                tour_nodes.append(curr_node)

                # update current node
                curr_node = curr_node.succ

                # remove current node from visited
                visited_nodes.remove(curr_node)

            # append last index to close the tour
            tour_nodes.append(curr_node)

            # check if there are still nodes in the visited nodes set
            # this will happen at unfeasible tours (like two subtours, instead of only one single tour)
            if len(visited_nodes) > 0:

                # get any unvisited node to continue the search
                curr_node = visited_nodes.pop()

        #  return the list of indexes
        return tour_nodes

    def shuffle(self):
        """
        Shuffle the tour nodes creating a random tour and re-initializing the tour edges 
        """

        # create random indexes between 1 and size-1
        indexes = [i for i in range(1, self.size)]
        shuffle(indexes)

        # add the 0 (first node) at the end of the indexes list
        indexes.append(0)

        # initilize the current node
        curr_node = self.nodes[0]

        # loop to change the predecessor, successor and position of each node
        for i in range(-1, self.size - 1):
            curr_node.succ = self.nodes[indexes[i + 1]]
            curr_node.pred = self.nodes[indexes[i - 1]]
            curr_node.pos = i + 1
            curr_node = curr_node.succ

        # rebuild the edges after shuffling the nodes
        self.set_edges()

    def restore(self, swaps=None):
        """
        Restore an initial tour that passed through some 2-opt swaps by doing a certain amount of reversed swaps using the swap_stack. If number of swaps is not defined, the entire swap stack will be undone

        :param swaps: the number of swaps to be undone
        :type swaps: int
        """

        # if swaps are not defined, the entire swap stack will be undone
        if swaps == None:
            swaps = len(self.swap_stack)

        # looping until number of swaps are reached
        for _ in range(swaps):

            # get last stack
            curr_stack = self.swap_stack[-1]

            # collect node values
            t1 = curr_stack[0]
            t2 = curr_stack[1]
            t3 = curr_stack[2]
            t4 = curr_stack[3]

            # collect the 2-Opt swap type made
            swap_type = curr_stack[-1]

            # execute the reversed swap based on the swap operation
            # swap is not recorded to the stack, since it is being undone
            if (swap_type == "swap_feasible"):
                self.swap_feasible(t4, t1, t2, t3, False, False)
            elif (swap_type == "swap_unfeasible"):
                self.swap_unfeasible(t4, t1, t2, t3, False, False)
            elif (swap_type == "swap_node_between_t2_t3"):
                self.swap_unfeasible(t4, t1, t2, t3, False, False)
            elif (swap_type == "swap_node_between_t2_t3_reversed"):
                self.swap_unfeasible(t4, t1, t2, t3, True, False)
            elif (swap_type == "swap_feasible_reversed"):
                self.swap_feasible(t4, t1, t2, t3, True, False)

            # remove last swap
            self.swap_stack.pop()

        # loop through each applied swap and check if there's any swap that don't recompute the pos attribute. If so, set the pos attribute
        for swap in self.swap_stack:
            if swap[-1] != "swap_feasible":
                self.set_pos()
                break

    def between(self, from_node, between_node, to_node, use_pos_attr=False):
        """
        Validate if a specific node is between two other nodes. There are two methods of search:

        1 - Using pos attribute: this is a fast search, but pos attribute must be correctly defined for the node segment being analyzed 
        2 - Using pred/succ attribute: a slow search (requires to traverse through the nodes), but doesn't require nodes to have pos attribute defined (this is relevant at unfeasible swaps)

        :param from_node: the starting node
        :type from_node: Node
        :param between_node: the node checked if between the other two nodes
        :type between_node: Node
        :param to_node: the ending node
        :type to_node: Node
        """

        # using pos attribute
        if use_pos_attr:

            # this is the check where the segment from-to does not contain the breaking pos value
            if from_node.pos <= to_node.pos:
                # segment does not pass through the breaking pos value
                return from_node.pos < between_node.pos and between_node.pos < to_node.pos
            else:
                # segment passes through the breaking pos value
                return from_node.pos < between_node.pos or between_node.pos < to_node.pos

        # using pred/succ attribute
        else:

            # get starting node
            node = from_node.succ

            # looping until to node is reached
            while node != to_node:
                # if between node is reached, return true
                if node == between_node:
                    return True
                else:
                    node = node.succ

            # if between node was not found, returns false
            return False

    def is_swap_feasible(self, t1, t2, t3, t4):
        """
        Validate if a 2Opt-Swap operation performed into a feasible tour/subtour results in another feasible tour/subtour using reference nodes. Edge (t1,t2) and (t3,t4) are broken so that a relink of (t2,t3) and (t1,t4) is made, as shown below:

          t4  t3           t4   t3
          ()--()           ()   ()
         /      \         /  \ /  \
        ()      ()  -->  ()   X   ()
         \      /         \  / \  /
          ()--()           ()   ()
          t2  t1           t2   t1

        For a feasible swap:
        *1: all nodes must be different from each other;
        *2: the order of the nodes in the tour must be (t1,t2,t4,t3) in any direction (clockwise or counter clockwise). Notice that in the scheme above, if we had (t1,t2,t3,t4), the relink would result in 2 closed loops, which are unfeasible. Notice that if we switch t1 with t2 and t3 with t4, it still feasible, the change was only made in the direction.

        * This function can be used into a subtour: when an unfeasible swap results in an unfeasible tour, there're two subtours that, isolated from each other, are valid subtours. This is relevant at the unfeasible search function (when node is between t1 and t4).

        :param t1: the tail node of the first broken edge
        :type t1: Node2D
        :param t2: the head node of the first broken edge
        :type t2: Node2D
        :param t3: the tail node of the second broken edge
        :type t3: Node2D
        :param t4: the head node of the second broken edge
        :type t4: Node2D
        :return: a boolean indicating if swap is valid and feasible
        :rtype: boolean
        """

        # for a feasible swap, all nodes must be different from each other (rule 1)
        if not (t1 != t2 and t1 != t3 and t1 != t4 and t2 != t3 and t2 != t4 and t3 != t4):
            return False

        # check t4 based on the orientation of the edge (t1,t2)
        # for a feasible swap, t4 can be validated using t1 and t2 orientation (rule 2)
        if t1.succ == t2:
            if t4 != t3.pred:
                return False
        elif t1.pred == t2:
            if t4 != t3.succ:
                return False

        # return true if swap is feasible
        return True

    def is_swap_unfeasible(self, t1, t2, t3, t4):
        """
        Validate if a 2Opt-Swap operation performed into a feasible tour results in an unfeasible tour using reference nodes. Edge (t1,t2) and (t3,t4) are broken so that a relink of (t2,t3) and (t1,t4) is made, as shown below:

          t3  t4           t3   t4
          ()--()           ()   ()
         /      \         / |   | \
        ()      ()  -->  () |   | ()
         \      /         \ |   | /
          ()--()           ()   ()
          t2  t1           t2   t1

        For an unfeasible swap:
        *1: all nodes must be different from each other;
        *2: the order of the nodes in the tour must be (t1,t2,t3,t4) in any direction (clockwise or counter clockwise), resulting in two separated subtours when relinking the edges;
        *3: each subtour must contain at least 3 nodes

        :param t1: the tail node of the first broken edge
        :type t1: Node2D
        :param t2: the head node of the first broken edge
        :type t2: Node2D
        :param t3: the tail node of the second broken edge
        :type t3: Node2D
        :param t4: the head node of the second broken edge
        :type t4: Node2D
        :return: a boolean indicating if swap is valid and unfeasible
        :rtype: boolean
        """

        # for an unfeasilbe swap, all nodes must be different from each other (rule 1)
        if not (t1 != t2 and t1 != t3 and t1 != t4 and t2 != t3 and t2 != t4 and t3 != t4):
            return False

        # check t4 based on the orientation of the edge (t1,t2)
        # for an unfeasible swap, t4 can be validated using t1 and t2 orientation (rule 2)
        if t1.succ == t2:
            if t4 == t3.pred:
                return False
        elif t1.pred == t2:
            if t4 == t3.succ:
                return False

        # t3 can't be a neighbor of t2 or t4 be a neighbor of t1: this result in a subtour segment with only 2 nodes, which is invalid (rule 3)
        if (t2.pred == t3 or t2.succ == t3 or t1.pred == t4 or t1.succ == t4):
            return False

        # return true if an unfeasible swap was found
        return True

    def is_swap_double_bridge(self, t1, t2, t3, t4, t5, t6, t7, t8):
        """
        Validate if a 4Opt-Swap double bridge operation performed into a feasible tour results in another feasible tour using reference nodes and, if a double bridge is found, returns the tuple of nodes in the sequence that the double bridge node shall be executed. Edges (t1,t2), (t3,t4), (t5,t6), (t7,t8) are broken so that a relink of nodes is made, as shown below:

            t3  t4             t3   t4
            ()--()             ()   ()
           /      \           / |   | \
        t8()      ()t5     t8()-------()t5 
          |        |   -->      |   |  
        t7()      ()t6     t7()-------()t6 
           \      /           \ |   | /
            ()--()             ()   ()
            t2  t1             t2   t1

        *It is possible to notice the double bridge move is a combination of two unfeasible swaps for segments t1-t4 and t5-t8. Both of this moves isolated results in two separated segments, while combined together results in a feasible tour.

        For an double bridge swap:
        *1: all nodes must be different from each other;
        *2: segment of 4 nodes t1-t4 and t5-t8 must result in an unfeasible swap 

        :param t1: the tail node of the first broken edge
        :type t1: Node2D
        :param t2: the head node of the first broken edge
        :type t2: Node2D
        :param t3: the tail node of the second broken edge
        :type t3: Node2D
        :param t4: the head node of the second broken edge
        :type t4: Node2D
        :param t5: the tail node of the third broken edge
        :type t5: Node2D
        :param t6: the head node of the third broken edge
        :type t6: Node2D
        :param t7: the tail node of the fourth broken edge
        :type t7: Node2D
        :param t8: the head node of the fourth broken edge
        :type t8: Node2D
        :return: a tuple of nodes for the double bridge swap (or empty if double bridge swap was not found)
        :rtype: tuple
        """

        # for a double bridge swap, all nodes must be different from each other (rule 1)
        if not (t1 != t3 and t1 != t4 and t1 != t5 and t1 != t6 and t1 != t7 and t1 != t8 and t2 != t3 and t2 != t4 and t2 != t5 and t2 != t6 and t2 != t7 and t2 != t8 and t3 != t5 and t3 != t6 and t3 != t7 and t3 != t8 and t4 != t5 and t4 != t6 and t4 != t7 and t4 != t8 and t5 != t7 and t5 != t8 and t6 != t7 and t6 != t8):
            return None

        # getting the nodes in succ sequence
        if t1.pred == t2:
            temp = t2
            t2 = t1
            t1 = temp
        if t3.pred == t4:
            temp = t4
            t4 = t3
            t3 = temp
        if t5.pred == t6:
            temp = t6
            t6 = t5
            t5 = temp
        if t7.pred == t8:
            temp = t8
            t8 = t7
            t7 = temp

        # get nodes in a sequence based on their position
        nodes = sorted((t1, t3, t5, t7), key=lambda el: el.pos)

        # return the nodes in the sequence for the double bridge swap
        return (nodes[0], nodes[0].succ, nodes[2], nodes[2].succ, nodes[1], nodes[1].succ, nodes[3], nodes[3].succ)

    def swap_feasible(self, t1, t2, t3, t4, is_subtour=False, record=True):
        """
        Execute a 2Opt-Swap operation in a feasible tour/subtour resulting in another feasible tour/subtour. Edge (t1,t2) and (t3,t4) are broken so that a relink of (t2,t3) and (t1,t4) is made, as shown below:

          t4  t3           t4   t3
          ()--()           ()   ()
         /      \         /  \ /  \
        ()      ()  -->  ()   X   ()
         \      /         \  / \  /
          ()--()           ()   ()
          t2  t1           t2   t1

        * This function can be used into a subtour: when an unfeasible swap results in an unfeasible tour, there're two subtours that, isolated from each other, are valid subtours. This is relevant at the unfeasible search function (when node is between t1 and t4). When this function is applied to subtours, the pos attribute must not be recomputed.

        :param t1: the tail node of the first broken edge
        :type t1: Node2D
        :param t2: the head node of the first broken edge
        :type t2: Node2D
        :param t3: the tail node of the second broken edge
        :type t3: Node2D
        :param t4: the head node of the second broken edge
        :type t4: Node2D
        :param is_subtour: a boolean indicating if swap is done into a subtour (from an unfeasible tour)
        :type is_subtour: boolean
        :param record: a boolean indicating if swap must be record into the swap stack
        :type record: boolean
        """

        # if t2 is not the successor of t1, invert t1 with t2 and t3 with t4
        # since the node reordering is always applied to t3 -> t1 segment, this is done so that the reordering loop will be done correctly.
        if (t1.succ != t2):
            temp = t1
            t1 = t2
            t2 = temp
            temp = t3
            t3 = t4
            t4 = temp

        # compute the segment size between t2 and t3 and compare it with the problem size
        # this validates if the segment between t2-t3 has more/less nodes than the segment t3-t1
        # if t2-t3 is shorter, we can switch t3 with t2 and t4 with t1 so that the swap result will be the same, but the amount of nodes to be reordered is smaller
        seg_size = t2.pos - t3.pos
        if seg_size < 0:
            seg_size += self.size  # the amount of nodes in t3-t1
        if (2 * seg_size > self.size):  # checking if there are more nodes in t3-t1 than t2-t3
            temp = t3
            t3 = t2
            t2 = temp
            temp = t4
            t4 = t1
            t1 = temp

        # initialize the position value from t1 (the value of t1 pos will be assigned to t3, since after the swap t3 is the predecessor of t2)
        pos = t1.pos

        # the starting node for the reversed segment is t3 (until t1)
        node = t3

        # the loop is broken at t1 successor, (since t1 is the last node)
        end_node = t1.succ

        # loop to reorder the nodes between t3-t1 segment (including t3 and t1)
        while (node != end_node):

            # invert the node ordering
            temp = node.succ
            node.succ = node.pred
            node.pred = temp

            # update node position only if is not a subtour
            if not is_subtour:
                node.pos = pos
                pos -= 1

            # update node for next loop with the last successor node
            node = temp

        # reassign the successor/predecessor values at each of the 4 reconnected nodes
        t3.succ = t2
        t2.pred = t3
        t1.pred = t4
        t4.succ = t1

        # update the swap stack
        if record:
            # the name is defined based on subtour parameter
            if not is_subtour:
                self.swap_stack.append((t1, t2, t3, t4, "swap_feasible"))
            else:
                self.swap_stack.append((t1, t2, t3, t4, "swap_feasible_reversed"))

    def swap_unfeasible(self, t1, t2, t3, t4, reverse_subtour=False, record=True):
        """
        Execute a 2Opt-Swap operation in a feasible tour resulting in an unfeasible tour. Edge (t1,t2) and (t3,t4) are broken so that a relink of (t2,t3) and (t1,t4) is made, as shown below:

          t3  t4           t3   t4
          ()--()           ()   ()
         /      \         / |   | \
        ()      ()  -->  () |   | ()
         \      /         \ |   | /
          ()--()           ()   ()
          t2  t1           t2   t1


        * Because some unfeasible swaps may reverse the subtour direction, the reverse_subtour parameter is used to reverse back the segment, useful when undoing those swaps.

        :param t1: the tail node of the first broken edge
        :type t1: Node2D
        :param t2: the head node of the first broken edge
        :type t2: Node2D
        :param t3: the tail node of the second broken edge
        :type t3: Node2D
        :param t4: the head node of the second broken edge
        :type t4: Node2D
        :param reverse_subtour: a boolean indicating if one sub-tour segment shall be reversed
        :type reverse_subtour: boolean
        :param record: a boolean indicating if swap must be record into the swap stack
        :type record: boolean
        """

        # reassign is done based on the direction of t1-t2 nodes
        if (t1.succ == t2):
            temp = t3
            t3 = t2
            t2 = temp
            temp = t4
            t4 = t1
            t1 = temp

        # reassign the successor/predecessor values at each of the 4 reconnected nodes
        t3.pred = t2
        t2.succ = t3
        t1.pred = t4
        t4.succ = t1

        # reverse the t1-t4 segment
        # relevant when undoing unfeasible swap of a node between t2 and t3, when the swap reverses one of the segments
        if reverse_subtour:

            # start from node t4
            node = t4

            # loop until t1
            while node.pred != t4:

                # reverse the node pred/succ attribute
                temp = node.pred
                node.pred = node.succ
                node.succ = temp

                # update for next node
                node = temp

            # reverse t1
            t1.pred = t1.succ
            t1.succ = t4

        # update the swap stack
        if record:
            self.swap_stack.append((t1, t2, t3, t4, "swap_unfeasible"))

    def swap_node_between_t2_t3(self, t1, t4, t5, t6, record=True):
        """
        Execute a 2Opt-Swap operation from an unfeasible tour with node t5 (and thus also t6) located between nodes t2 and t3. The break of edge (t1,t4) opens one of the closed tour and  connecting t1-t6 and t4-t5 results into a new valid tour.

        This function is written with reference nodes t5 and t6, but it works at any swap level if starting from an unfeasible tour. As an example, this function is used in LK algorithm to perform the swap of t7 (and thus also t8) between t2-t3.

        :param t1: the tail node of the first broken edge
        :type t1: Node2D
        :param t4: the head node of the first broken edge
        :type t4: Node2D
        :param t5: the tail node of the second broken edge
        :type t5: Node2D
        :param t6: the head node of the second broken edge
        :type t6: Node2D
        :param record: a boolean indicating if swap must be record into the swap stack
        :type record: boolean
        """

        # checking if t1-t4 is reversed (t4 = t1.pred) or not (t4 = t1.succ)
        # this is required to check succ/pred sequences
        t4_after_t1 = t1.succ == t4

        # checking if t5-t6 is reversed (t6 = t5.pred) or not (t6 = t5.succ)
        # this is required to check succ/pred sequences
        t6_after_t5 = t5.succ == t6

        # a boolean checking if segment must be reversed
        # requires to reverse t5-t6 segment when (t1-t4-t5-t6) or (t4-t1-t6-t5)
        # otherwise the tour (luckly) is in the correct sequence
        reverse_subtour = t4_after_t1 != t6_after_t5

        # apply the reverse loop
        if (reverse_subtour):

            # arranging the nodes for the reorder loop (t5-t6 segment)
            # reorder is always going backwards (using pred attribute)
            from_node = t6
            to_node = t5
            if t6_after_t5:
                from_node = t5
                to_node = t6

            # loop reversing the segment
            while (from_node != to_node):

                # reverse the node pred/succ attribute and update for next node
                temp = from_node.pred
                from_node.pred = from_node.succ
                from_node.succ = temp
                from_node = temp

            # reverse the last node
            temp = to_node.pred
            to_node.pred = to_node.succ
            to_node.succ = temp

        # reassign the swap nodes based on tour direction
        if (t4_after_t1):
            t1.succ = t6
            t6.pred = t1
            t5.succ = t4
            t4.pred = t5
        else:
            t1.pred = t6
            t6.succ = t1
            t5.pred = t4
            t4.succ = t5

        # update the swap stack
        if record:
            # record also in the swap name if the reversed loop was applied or not
            # this is relevant when undoing the swap
            if reverse_subtour:
                self.swap_stack.append((t1, t4, t5, t6, "swap_node_between_t2_t3_reversed"))
            else:
                self.swap_stack.append((t1, t4, t5, t6, "swap_node_between_t2_t3"))

    def swap_double_bridge(self, t1, t2, t3, t4, t5, t6, t7, t8, record=True):
        """
        Execute a 4Opt-Swap double bridge operation performed into a feasible tour resulting in another feasible tour using reference nodes. Edges (t1,t2), (t3,t4), (t5,t6), (t7,t8) are broken so that a relink of nodes is made, as shown below:

            t3  t4             t3   t4
            ()--()             ()   ()
           /      \           / |   | \
        t8()      ()t5     t8()-------()t5 
          |        |   -->      |   |  
        t7()      ()t6     t7()-------()t6 
           \      /           \ |   | /
            ()--()             ()   ()
            t2  t1             t2   t1

        :param t1: the tail node of the first broken edge
        :type t1: Node2D
        :param t2: the head node of the first broken edge
        :type t2: Node2D
        :param t3: the tail node of the second broken edge
        :type t3: Node2D
        :param t4: the head node of the second broken edge
        :type t4: Node2D
        :param t5: the tail node of the third broken edge
        :type t5: Node2D
        :param t6: the head node of the third broken edge
        :type t6: Node2D
        :param t7: the tail node of the fourth broken edge
        :type t7: Node2D
        :param t8: the head node of the fourth broken edge
        :type t8: Node2D
        :param record: a boolean indicating if swap must be record into the swap stack
        :type record: boolean
        """

        # execute the first unfeasible swap
        self.swap_unfeasible(t1, t2, t3, t4, False, False)

        # the second unfeasible swap is done in such way that new tour does not require segment reordering
        # checking direction of initial tour segment t1-t4
        from_node = t4
        to_node = t1
        if t1.pred == t2:
            from_node = t1
            to_node = t4

        # check if t5 is between t4 and t1 segment
        # if not, reorder the nodes so that t5 is between t1-t4
        if not self.between(from_node, t5, to_node):
            temp = t5
            t5 = t8
            t8 = temp
            temp = t6
            t6 = t7
            t7 = temp

        # checking if t5-t6 and t7-t8 needs to be switched to match tour orientation
        if (t1.succ == t2 and t5.pred == t6) or (t1.pred == t2 and t5.succ == t6):
            temp = t5
            t5 = t6
            t6 = t5
            temp = t7
            t7 = t8
            t8 = temp

        # after reordering the nodes, the second unfeasible swap is performed
        self.swap_unfeasible(t5, t6, t7, t8, False, False)

        # update pos attribute
        self.set_pos()

        # update the swap stack
        if record:
            self.swap_stack.append((t1, t2, t3, t4, t5, t6, t7, t8, "swap_double_bridge"))

    def __str__(self):
        """
        The display string when printing the object

        :return: the display string
        :rtype: str
        """

        # get the starting node
        curr_node = self.nodes[0]

        # initialize the string
        node_seq = str(curr_node.id)

        # loop until tour is closed
        while curr_node.succ != self.nodes[0]:

            # get next node
            curr_node = curr_node.succ

            # update the string
            node_seq += f",{curr_node.id}"

        # return the string sequence
        return f"({node_seq})"



import math
import logging
import random
from itertools import permutations
from lk_heuristic.models.edge import Edge
from lk_heuristic.models.tour import Tour
from lk_heuristic.models.node import NodePivot


class Tsp:
    """
    The TSP class represents the travelling salesman problem inputs and helper methods for the problem.
    """

    # set the precision of the gain
    # this is required when computing the gain in "simmetric" tours, to avoid incorrect calculations when using just "> 0"
    gain_precision = 0.01

    def __init__(self, nodes, cost_function, shuffle=False, backtracking=(5, 5), reduction_level=4, reduction_cycle=4, tour_type="cycle", logging_level=logging.INFO):
        """
        The TSP input is a list of nodes which will be used as input to build a tour and a cost function to build the cost matrix.

        :param nodes: the list of nodes for the tsp
        :type nodes: list
        :param cost_function: the function used to compute the cost between two nodes
        :type cost_function: function
        :param shuffle: a boolean value indicating if nodes shall be shuffled before building the tour
        :type shuffle: boolean
        :param backtracking: the number of closest neighbors to use in each backtracking level.
        :type backtracking: tuple
        :param reduction_level: the starting level in current optimization cycle where reducted edges will start being considered
        :type reduction_level: int
        :param reduction_cycle: the number of optimization cycles when reducted edges will start being considered
        :type reduction_cycle: int
        :param tour_type: the type of the tour (either 'path' or 'cycle')
        :type tour_type: str
        :param logging_level: the level for logging messages
        :type logging_level: int
        """

        # setup the logger
        self.logger = logging.getLogger(__name__)
        self.logger.setLevel(logging_level)

        # initialize nodes
        self.nodes = nodes

        # initialize the tour type
        self.tour_type = tour_type

        # initialize the tour using the nodes and the tour type
        self.tour = Tour(self.nodes, t=tour_type)

        # shuffle the tour nodes
        if shuffle:
            self.tour.shuffle()

        # a boolean indicating if tour nodes shall be shuffled at every restart of the optimization loop
        # shuffle will not be applied when a double bridge move is found using lk1 optimizer
        self.shuffle = True

        # initialize the cost matrix after having the nodes initialized into the tour object using the defined cost_function
        self.cost_matrix = {}
        self.set_cost_matrix(cost_function)

        # initialize tour cost after having the cost matrix initialized
        self.tour.set_cost(self.cost_matrix)

        # solution methods avaliable to solve the problem
        self.methods = {
            "bf_improve": self.bf_improve,
            "nn_improve": self.nn_improve,
            'lk1_improve': self.lk1_improve,
            "lk2_improve": self.lk2_improve,
        }

        # init the set of solutions
        # the set of solutions will help to avoid repeated tours analysis
        # it is a hash of a tuple of node sequences
        self.solutions = set()

        # init a dict of closest neighbors
        # each node will have a maximum number of closest neighbors defined by the cost between them. It is required to use the maximum value of the backtracking.
        self.closest_neighbors = {}
        self.set_closest_neighboors(max_neighbors=max(backtracking))

        # init the bactracking levels
        # each value represent the number of neighbors to try in each search level
        self.backtracking = backtracking

        # init the reduction starting level, starting cycle and reduction edges set
        # this set contains the non-intersected edges at a maximum reduction levels
        # the set is initialized empty and populated after first local optima is found
        self.reduction_level = reduction_level
        self.reduction_cycle = reduction_cycle
        self.reduction_edges = set()

        # init the improvement cycles count, updated everytime a local optima is found
        # this parameter is only relevant when running many improvement cycles, used to apply reduction and double bridge moves
        self.cycles = 0

        # gains values collected during the search, the best gain value and the double bridge gain
        self.close_gains = []
        self.best_close_gain = 0
        self.double_bridge_gain = 0

    def set_cost_matrix(self, cost_func):
        """
        Compute the cost matrix using tsp nodes and a predefined cost function. The matrix is in the form of a dictionary with (row,column) values as keys of the dict

        :param cost_func: the cost function to be used to compute the cost matrix
        :type cost_func: function
        """

        # index is used to avoid repeated computation
        index = 0
        for i in range(len(self.nodes)):
            for j in range(index, len(self.nodes)):
                # get node 1 and node2
                n1 = self.nodes[i]
                n2 = self.nodes[j]

                # compute the cost between nodes
                # for hamiltonian path, pivot nodes are used and edges containing those nodes will have zero cost
                cost = 0
                if not (type(n1) == NodePivot or type(n2) == NodePivot):
                    cost = cost_func(n1, n2)

                # update cost matrix
                self.cost_matrix[(n1.id, n2.id)] = cost
                self.cost_matrix[(n2.id, n1.id)] = cost  # simmetric value

            # update index for next iteration
            index += 1

    def set_closest_neighboors(self, max_neighbors):
        """
        Get the closest neighbors of each tsp node using the cost_matrix values. The return will be a dict of each node as key and the value is a list of closest nodes. The number of values to be returned inside the list is defined by the max_neighboors variable.

        This function, combined with the get_best_neighbors, is the lookahead refinement from LK paper, where the best swaps from the closest neighbors of a node direct the search and speed up performance.

        :param max_neighboors: the maximum number of closest neighbors to collect
        :type max_neighboors: int
        :return: returns a dict mapping each node with its closest neighbors
        :rtype: list
        """

        # loop through each node
        for node1 in self.tour.nodes:

            # get all edge costs for each possible neighbor
            neighbors = [(node2, self.cost_matrix[(node1.id, node2.id)]) for node2 in self.tour.nodes if node2 != node1]

            # sort the neighbors based on the cost value and get the smallest ones
            neighbors_min = sorted(neighbors, key=lambda x: x[1])[:max_neighbors]

            # populate neighbor list with the best neighbors
            self.closest_neighbors[node1] = [neighbor[0] for neighbor in neighbors_min]

    def get_best_neighbors(self, t2, t1=None):
        """
        Get the best tuple of nodes (t3,t4) by computing the gain of swapping the closest neighbors of t2 (i.e, swapping (t3,t4) with (t2,t3)). This is the lookahead refinement where the nodes to be selected for executing the swap are sorted/selected by best gain. Node t1 is used in this function to check if the swap (t1,t2,t3,t4) is valid. If t1 is not provided, the swap is done without checking if it is valid.

        :param t2: the t2 node from where best neighbors shall be found
        :type t2: Node
        :param t1: the t1 node (neighbor of t2 that makes the broken edge)
        :type t1: Node
        :return: a dict with keys (t3,t4) and value is the gain
        :rtype: dict
        """

        # a dict storing the best neighbors from t2
        best_neighbors = {}

        # loop through each closest neighbor of t2 (since looping throgh every neighbor would be time consuming)
        for t3 in self.closest_neighbors[t2]:
            # loop through each possible t4
            for t4 in (t3.pred, t3.succ):
                # check if swap is valid
                if (t1):
                    if (self.tour.is_swap_feasible(t1, t2, t3, t4)):
                        # compute the gain
                        best_neighbors[(t3, t4)] = self.cost_matrix[(t3.id, t4.id)] - self.cost_matrix[t2.id, t3.id]
                # compute the gain not checking for valid swaps
                else:
                    # compute the gain
                    best_neighbors[(t3, t4)] = self.cost_matrix[(t3.id, t4.id)] - self.cost_matrix[t2.id, t3.id]

        # returns a list of (key,value) pairs of the max values of gain.
        return sorted(best_neighbors.items(), key=lambda x: x[1], reverse=True)

    def lk1_feasible_search(self, level, gain, swap_func, t1, t2, t3, t4, broken_edges, joined_edges):
        """
        This is the main search loop of LK Heuristic, trying to find broken and joined edges in such way that a feasible tour with lower cost is found. The search is recursivelly called while potential nodes exists. 

        The search store both the gain value of closing the loop and the swap memory of the executed swaps. This memory is used when exiting the recursive call, so that is possible to select the best gain found in the search by undoing the executed swaps until the best gain position.

        This function may be called from the unfeasible swap function. In this case, the starting swap function will be a special function that turns an unfeasible tour into a feasible one. 

        :param level: the current level of the search to compare with backtracking levels and reduction level
        :type level: int
        :param gain: the current gain from last LK step
        :type gain: float
        :param swap_func: the swap function to be used at nodes (t1, t2, t3, t4)
        :type swap_func: str
        :param t1: the tail node of the first broken edge in the 2-opt swap
        :type t1: Node2D
        :param t2: the head node of the first broken edge in the 2-opt swap
        :type t2: Node2D
        :param t3: the tail node of the second broken edge in the 2-opt swap
        :type t3: Node2D
        :param t4: the head node of the second broken edge in the 2-opt swap
        :type t4: Node2D
        :param broken_edges: the set of broken edges
        :type broken_edges: set
        :param joined_edges: the set of joined edges
        :type joined_edges: set
        """

        # step 4(a) at LK Paper is omitted here since the feasibility criterion is done when searching for the best neighbors in a swap. When entering this function, step 4(a) was already checked by the get best neighbors function.

        # set xi edge (the new edge to be broken)
        broken_edge = Edge(t3, t4)
        broken_cost = self.cost_matrix[(t3.id, t4.id)]

        # apply the reduction refinement from LK Paper
        # reduction is not applied in current optimization cycle if level is greater than a certain level
        # reduction is not applied if a certain amount of optimization cycles were not run yet
        if (level >= self.reduction_level and self.cycles <= self.reduction_cycle and broken_edge in self.reduction_edges):
            return

        # update the sets of broken and joined edges using the swap nodes
        broken_edges.add(Edge(t1, t2))
        joined_edges.add(Edge(t2, t3))

        # execute the swap based on the swap function to be executed
        # when coming from previous feasible swap, a feasible swap is reaplied
        # when coming from a previous unfeasible swap, the node between t2 and t3 is applied
        if swap_func == "swap_feasible":
            self.tour.swap_feasible(t1, t2, t3, t4)
        elif swap_func == "swap_node_between_t2_t3":
            self.tour.swap_node_between_t2_t3(t1, t2, t3, t4)

        # set yi edge to close the tour (instead of continuing exploration)
        # also check that close joined edge is valid (disjoint and not already in tour)
        joined_close_edge = Edge(t4, t1)
        joined_close_cost = self.cost_matrix[(t4.id, t1.id)]
        joined_close_valid = joined_close_edge not in self.tour.edges and joined_close_edge not in broken_edges

        # compute the gain of closing the tour and add it to close gain list
        # also update the best gain found so far for closing the tour
        close_gain = gain + (broken_cost - joined_close_cost)
        self.close_gains.append(close_gain)
        self.best_close_gain = close_gain if close_gain > self.best_close_gain else self.best_close_gain

        # get the number of backtracked neighbors at current level (defaults to 1 if no value is defined at backtracking parameter)
        curr_backtracking = 1
        if (level <= len(self.backtracking) - 1):
            curr_backtracking = self.backtracking[level]

        # yi is selected based on best neighbors of t4
        # next_y_head is the selected yi head node (at the end of the yi)
        # next_x_head is the xi+1 head node (at the end of the xi+1)
        # this is step 4(b) at LK Paper (with the lookahead refinement - 2.B)
        for (next_y_head, next_x_head), _ in self.get_best_neighbors(t4, t1)[:curr_backtracking]:

            # set yi edge
            joined_edge = Edge(t4, next_y_head)
            joined_cost = self.cost_matrix[(t4.id, next_y_head.id)]

            # compute gain for exploration (i.e, if not closing the tour)
            explore_gain = gain + (broken_cost - joined_cost)

            # disjoint criteria (xi can't be previously joined and yi can't be previously broken)
            # it is also required to check if broken edge is not repeated and if joined edge is not already in tour
            # this is step 4(c) in LK Paper
            disjoint_criteria = False
            if broken_edge not in broken_edges and broken_edge not in joined_edges:
                if joined_edge not in self.tour.edges and joined_edge not in broken_edges:
                    disjoint_criteria = True

            # gain criteria (gain must be positive)
            # this is step 4(d) in LK Paper
            gain_criteria = False
            if explore_gain > self.gain_precision:
                gain_criteria = True

            # xi+1 criteria (next x must be possible to be broken)
            # next x can be broken if it follows the disjoint criteria and is not repeated (already broken)
            # this is step 4(e) in LK Paper
            next_xi_criteria = False
            next_broken_edge = Edge(next_y_head, next_x_head)
            if (next_broken_edge not in broken_edges and next_broken_edge not in joined_edges):
                next_xi_criteria = True

            # checking the 3 required criterias (4c, 4d and 4e), as mentioned in step 4(b) in LK Paper
            if disjoint_criteria and gain_criteria and next_xi_criteria:

                # check for repeated tours (checkout refinement - 2.A)
                if (hash(tuple([node.succ.id for node in self.tour.nodes])) in self.solutions):
                    return

                # checking if closing the tour will lead to a better gain than continuing exploration
                # if close is better than explore, ends the loop
                # if explore is better than close, call the search loop with new nodes
                # this is step 4(f) at LK Paper
                if (close_gain > explore_gain and close_gain >= self.best_close_gain and close_gain > self.gain_precision and joined_close_valid):

                    # update the sets of broken and joined edges
                    broken_edges.add(broken_edge)
                    joined_edges.add(joined_close_edge)

                    # end the loop if closing is better than exploring
                    return

                else:

                    # continue the exploration, if it is better than closing the tour
                    # this is the return to step 4 with i = i + 1 at LK Paper
                    return self.lk1_feasible_search(level + 1, explore_gain, "swap_feasible", t1, t4, next_y_head, next_x_head, broken_edges, joined_edges)

    def lk1_unfeasible_search(self, gain, t1, t2, t3, t4, broken_edges, joined_edges):
        """
        This is the alternative search loop of LK Heuristic at level 1, when the swap of the nodes t1, t2, t3, t4 leads to an unfeasible tour. LK suggest that in this condition, a special search shall be made looking for nodes t5, t6 (t7 and t8, if required) in such way that the unfeasible tour turns into a feasible one. If a feasible tour is found with better gain, a call for the main loop search is made. This is the step 6(b) in LK Paper.

        :param gain: the current gain from last LK step
        :type gain: float
        :param t1: the tail node of the first broken edge in the 2-opt swap
        :type t1: Node2D
        :param t2: the head node of the first broken edge in the 2-opt swap
        :type t2: Node2D
        :param t3: the tail node of the second broken edge in the 2-opt swap
        :type t3: Node2D
        :param t4: the head node of the second broken edge in the 2-opt swap
        :type t4: Node2D
        :param broken_edges: the set of broken edges
        :type broken_edges: set
        :param joined_edges: the set of joined edges
        :type joined_edges: set
        """

        # update the sets of broken and joined edges using the swap nodes
        # adding x1 and y1
        broken_edges.add(Edge(t1, t2))
        joined_edges.add(Edge(t2, t3))

        # the broken edge (x2) that will lead to 2 separated tours
        broken_edge_1 = Edge(t3, t4)
        broken_cost_1 = self.cost_matrix[(t3.id, t4.id)]

        # execute the unfeasible swap (creating two separated tours)
        # append a dummy gain value for the unfeasible swap
        # this is the step represented at Fig. 4(a) in LK Paper
        self.tour.swap_unfeasible(t1, t2, t3, t4)
        self.close_gains.append(-1)

        # get the backtracking at second level
        curr_backtracking = 1
        if (len(self.backtracking) - 1 >= 1):
            curr_backtracking = self.backtracking[1]

        # looking for the best neighbors of t4 at second level of backtracking
        for (t5, t6), _ in self.get_best_neighbors(t4)[:curr_backtracking]:

            # set y2 edge
            joined_edge_1 = Edge(t4, t5)
            joined_cost_1 = self.cost_matrix[(t4.id, t5.id)]

            # compute gain for exploration (not closing the tour but accepting the exploring node t5)
            explore_gain = gain + (broken_cost_1 - joined_cost_1)

            # gain criteria (gain must be positive)
            # the gain criteria shall be validated even if using an unfeasible tour
            gain_criteria = False
            if explore_gain > self.gain_precision:
                gain_criteria = True

            # checking if (t5,t6) is a valid choice
            # t5 and t6 must be different from t1,t2,t3,t4 (it must lie between those nodes)
            valid_nodes = False
            if (t5 != t1 and t5 != t2 and t5 != t3 and t5 != t4):
                if (t6 != t1 and t6 != t2 and t6 != t3 and t6 != t4):
                    valid_nodes = True

            # checking for valid node selection with positive gain
            if valid_nodes and gain_criteria:

                # set x3 edge
                broken_edge_2 = Edge(t5, t6)
                broken_cost_2 = self.cost_matrix[(t5.id, t6.id)]

                # a boolean checking if t5 is between t1-t4 segment
                t5_between_t1_t4 = False

                # checking the order of the segment t1-t4
                t1_after_t4 = t4.succ == t1

                # validate if t5 is between t1-t4
                if t1_after_t4:
                    t5_between_t1_t4 = self.tour.between(t1, t5, t4)
                else:
                    t5_between_t1_t4 = self.tour.between(t4, t5, t1)

                # t5 condition of being between t1 and t4
                if (t5_between_t1_t4):

                    # checking if t5 swap is feasible
                    if (self.tour.is_swap_feasible(t1, t4, t5, t6)):

                        # get the backtracking at third level
                        curr_backtracking = 1
                        if (len(self.backtracking) - 1 >= 2):
                            curr_backtracking = self.backtracking[2]

                        # looking for (t7,t8)
                        for (t7, t8), _ in self.get_best_neighbors(t6)[:curr_backtracking]:

                            # set y3 edge
                            joined_edge_2 = Edge(t6, t7)
                            joined_cost_2 = self.cost_matrix[(t6.id, t7.id)]

                            # update the exploration gain
                            explore_gain += (broken_cost_2 - joined_cost_2)

                            # gain criteria (gain must be positive)
                            gain_criteria = False
                            if explore_gain > self.gain_precision:
                                gain_criteria = True

                            # a boolean checking if t7 is between t2-t3 segment
                            t7_between_t2_t3 = False

                            # checking the order of the segment t2-t3
                            t2_after_t3 = t3.succ == t2

                            # validate if t7 is between t2-t3
                            if t2_after_t3:
                                t7_between_t2_t3 = self.tour.between(t2, t7, t3)
                            else:
                                t7_between_t2_t3 = self.tour.between(t3, t7, t2)

                            # checking if (t7,t8) is a valid choice
                            # t7 and t8 must be different from t2,t3 (it must lie between those nodes)
                            valid_nodes = False
                            if (t7 != t2 and t7 != t3 and t8 != t2 and t8 != t3):
                                valid_nodes = True

                            # (t7,t8) will define the x4 broken edge. LK suggest to select the greater x4 value possible from t7. This is something already done by the get_best_neighbors function, since it selected the maximum delta between (t8,t7) - (t6,t7)

                            # validating gain criteria and if selected nodes are valid
                            if gain_criteria and valid_nodes and t7_between_t2_t3:

                                # set xi edge (the new edge to be broken)
                                broken_edge_3 = Edge(t7, t8)

                                # execute the swap of t5 being between t1-t4 and append a dummy value
                                # for this special swap, t1-t4 is a subtour of an unfeasible tour
                                self.tour.swap_feasible(t1, t4, t5, t6, is_subtour=True)
                                self.close_gains.append(-1)

                                # update the sets of broken and joined edges
                                broken_edges.add(broken_edge_1)
                                broken_edges.add(broken_edge_2)
                                broken_edges.add(broken_edge_3)
                                joined_edges.add(joined_edge_1)
                                joined_edges.add(joined_edge_2)

                                # continue the exploration with feasible search, applied with a special swap function
                                # this is the return to step 4 with i = i + 1 at LK Paper
                                return self.lk1_feasible_search(4, explore_gain, "swap_node_between_t2_t3", t1, t6, t7, t8, broken_edges, joined_edges)

                # if t5 is between t2 and t3, t6 will be joined to t1 to close the loop
                else:

                    # update the sets of broken and joined edges
                    broken_edges.add(broken_edge_1)
                    broken_edges.add(broken_edge_2)
                    joined_edges.add(joined_edge_1)

                    # continue the exploration, if it is better than closing the tour
                    # this is the return to step 4 with i = i + 1 at LK Paper
                    return self.lk1_feasible_search(3, explore_gain, "swap_node_between_t2_t3", t1, t4, t5, t6, broken_edges, joined_edges)

        # if no improvement was found in the unfeasible search, undo the unfeasible swap
        self.tour.restore()

    def lk1_double_bridge_search(self, max_tests=100):
        """
        Execute a random search for a double bridge move that is feasible and results in a better tour. If this double bridge move is found, the move is executed, tour edges are updated and the double bridge gain for this move is stored into class variable

        :param max_tests: the maximum number of double bridge tests to be made
        :type max_tests: int
        :return: the execution of the double bridge move
        :rtype: None
        """

        # getting the edges that can be broken (using only edges not in reduction set)
        search_edges = list(self.tour.edges.difference(self.reduction_edges))

        # the double bridge swap requires 4 edges
        if len(search_edges) >= 4:

            # loop until max number of tests are reached
            for _ in range(max_tests):

                # shuffle the edges
                random.shuffle(search_edges)

                # get the broken edges
                broken_edge_1 = search_edges[0]
                broken_edge_2 = search_edges[1]
                broken_edge_3 = search_edges[2]
                broken_edge_4 = search_edges[3]

                # try to get the double bridge swap configuration
                double_bridge_nodes = self.tour.is_swap_double_bridge(broken_edge_1.n1, broken_edge_1.n2, broken_edge_2.n1, broken_edge_2.n2, broken_edge_3.n1, broken_edge_3.n2, broken_edge_4.n1, broken_edge_4.n2)

                # if configuration is found continue the double bridge move
                if double_bridge_nodes:

                    # get all nodes of the swap operation
                    t1 = double_bridge_nodes[0]
                    t2 = double_bridge_nodes[1]
                    t3 = double_bridge_nodes[2]
                    t4 = double_bridge_nodes[3]
                    t5 = double_bridge_nodes[4]
                    t6 = double_bridge_nodes[5]
                    t7 = double_bridge_nodes[6]
                    t8 = double_bridge_nodes[7]

                    # compute the broken costs
                    broken_cost_1 = self.cost_matrix[(t1.id, t2.id)]
                    broken_cost_2 = self.cost_matrix[(t3.id, t4.id)]
                    broken_cost_3 = self.cost_matrix[(t5.id, t6.id)]
                    broken_cost_4 = self.cost_matrix[(t7.id, t8.id)]

                    # compute the joined costs
                    joined_cost_1 = self.cost_matrix[(t1.id, t4.id)]
                    joined_cost_2 = self.cost_matrix[(t2.id, t3.id)]
                    joined_cost_3 = self.cost_matrix[(t5.id, t8.id)]
                    joined_cost_4 = self.cost_matrix[(t6.id, t7.id)]

                    # compute the gain
                    gain = (broken_cost_1 + broken_cost_2 + broken_cost_3 + broken_cost_4) - (joined_cost_1 + joined_cost_2 + joined_cost_3 + joined_cost_4)

                    # checking if the gain is positive
                    if (gain > self.gain_precision):

                        # execute double bridge swap
                        self.tour.swap_double_bridge(t1, t2, t3, t4, t5, t6, t7, t8)

                        # update the double bridge gain
                        self.double_bridge_gain = gain

                        # update tour edges
                        self.tour.edges.remove(Edge(t1, t2))
                        self.tour.edges.remove(Edge(t3, t4))
                        self.tour.edges.remove(Edge(t5, t6))
                        self.tour.edges.remove(Edge(t7, t8))
                        self.tour.edges.add(Edge(t1, t4))
                        self.tour.edges.add(Edge(t2, t3))
                        self.tour.edges.add(Edge(t5, t8))
                        self.tour.edges.add(Edge(t6, t7))

                        # break the loop
                        break

    def lk1_main(self):
        """
        Execute the main loop of LK Heuristic to optimize the tour. This is a single run of LK loop, which tries to find a better tour considering current tour being analyzed.

        :return: a boolean value indicating if an improve was found
        :rtype: boolean
        """

        # loop through each node in the tour, to apply the optimization
        # LK Heuristic will loop over all tour nodes as initial nodes (first node is explored in total)
        # this is step 2 and 6(e) in LK Paper
        for t1 in self.tour.get_nodes():

            # loop through each neighboorhood node from initial node (second node is also explored as total, so at least 2-opt is always tested for all nodes)
            # this is step 2 and 6(d) in LK Paper
            for t2 in (t1.pred, t1.succ):

                # get the break edge (x1) and the cost
                # this is step 2 in LK Paper
                broken_edge = Edge(t1, t2)
                broken_cost = self.cost_matrix[(t1.id, t2.id)]

                # loop through the best possible nodes (t3,t4) from t2 instead of looping through all nodes. The number of best nodes is defined by the backtracking parameter.
                # this is step 3 and 6(c) in LK Paper (with additional lookahead refinement - 2.B in LK paper)
                # there's also step 4(a) in the get best neighbors, since the function checks if swap is valid for the nodes t1,t2,t3,t4
                for (t3, t4), _ in self.get_best_neighbors(t2)[:self.backtracking[0]]:

                    # get the joined edge (y1) and its cost
                    joined_edge = Edge(t3, t2)
                    joined_cost = self.cost_matrix[(t3.id, t2.id)]

                    # compute the gain
                    gain = broken_cost - joined_cost

                    # check if joined edge is not inside the existing tour
                    # check if gain is positive
                    # this is step 3 in LK Paper
                    if joined_edge not in self.tour.edges and gain > self.gain_precision:

                        # initialize broken and joined edges set
                        broken_edges = set()
                        joined_edges = set()

                        # execute the search loop
                        # if swap is feasible, execute the normal search
                        # if swap is unfeasible (2 separated tours), execute specific search from step 6(b) in LK Paper
                        if (self.tour.is_swap_feasible(t1, t2, t3, t4)):
                            self.lk1_feasible_search(1, gain, "swap_feasible", t1, t2, t3, t4, broken_edges, joined_edges)
                        elif (self.tour.is_swap_unfeasible(t1, t2, t3, t4)):
                            self.lk1_unfeasible_search(gain, t1, t2, t3, t4, broken_edges, joined_edges)

                        # check if a new gain was found
                        if self.close_gains:

                            # check if best gain is positive
                            # this is step 5 of LK Paper
                            if (max(self.close_gains) > 0):

                                # get the index of the best gain found
                                # LK search is done until gain is positive, but the best gain may not be the last one found
                                best_index = self.close_gains.index(max(self.close_gains))

                                # update tour edges joined and removed until best swap (which matches the index of the best gain)
                                for i in range(best_index + 1):
                                    (n1, n2, n3, n4, _) = self.tour.swap_stack[i]
                                    self.tour.edges.remove(Edge(n1, n2))
                                    self.tour.edges.remove(Edge(n3, n4))
                                    self.tour.edges.add(Edge(n2, n3))
                                    self.tour.edges.add(Edge(n4, n1))

                                # undo executed swaps until best gain index
                                self.tour.restore((len(self.close_gains) - 1) - best_index)

                                # compute the new cost for the new tour
                                self.tour.set_cost(self.cost_matrix)

                                # reset close gains
                                self.close_gains.clear()

                                # return true if improvement was found
                                return True

                            else:

                                # reset close gains
                                self.close_gains.clear()

                                # if best gain found during search is negative, restore the initial tour (since no improvement was found, but swaps were made)
                                self.tour.restore()

        # if no improvement is found, return false
        return False

    def lk1_improve(self):
        """
        The improve loop using Lin-Kernighan Heuristic. The loop calls main LK optimizer everytime a better tour is found (which will try to optimize the new best tour using the same optimizer). The loop ends when no improvement is found, such that current tour will be the local optimal.

        After a local optimal is found, the preparation for 2 refinements are done:
        *for reduction refinement, edges of new tour is intersected with last tour edges. The intersection is used for the refinement during the main improvement loop.
        *for double-bridge refinement, after a certain amount of improvement cycles are run, the non-reduction edges are used to perform a search for a double-bridge move
        """

        # number of new tours achieved
        tour_count = 1

        # boolean checking if a improved tour was found
        improved = True

        # log the starting tour cost
        self.logger.debug(f"Starting tour cost: {self.tour.cost:.3f}")

        # loop until no improvement is found at TSP tour
        # this is the step 5 in LK Paper
        while improved:

            # compute the improvement
            improved = self.lk1_main()

            # get executed swaps count
            total_swaps = len(self.tour.swap_stack)
            feasible_swaps = len([swap for swap in self.tour.swap_stack if swap[-1] == "swap_feasible"])
            unfeasible_swaps = len([swap for swap in self.tour.swap_stack if swap[-1] != "swap_feasible"])

            # log the current tour cost
            self.logger.debug(f"Current tour '{tour_count}' cost: {self.tour.cost:.3f} / gain: {self.best_close_gain:.3f} / swaps: {total_swaps} / feasible swaps: {feasible_swaps} / unfeasible swaps: {unfeasible_swaps}")

            # update tour count
            tour_count += 1

            # reset tsp parameters for next iteration
            self.tour.swap_stack.clear()
            self.best_close_gain = 0

        # update improvement cycle count
        self.cycles += 1

        # add the current solution string to the set (direct and reversed order)
        self.solutions.add(hash(tuple([node.succ.id for node in self.tour.nodes])))
        self.solutions.add(hash(tuple([node.pred.id for node in self.tour.nodes])))

        # initialize or update the reduction edges by intersecting reduction edges with new tour edges
        self.reduction_edges = set(self.tour.edges) if self.cycles == 1 else self.reduction_edges.intersection(self.tour.edges)

        # execute the double bridge refinement after certain amount of optimization cycles
        if self.cycles >= self.reduction_cycle:

            # search for a double bridge move
            self.lk1_double_bridge_search()

            # check if a good double bridge move was found
            if (self.double_bridge_gain > 0):

                # log the current tour cost
                self.logger.info(
                    f"Double bridge move found: cost: {self.tour.cost:.3f} / gain: {self.double_bridge_gain:.3f}")

                # reset the double brige gain
                self.double_bridge_gain = 0

                # reset the shuffle value (for next iteration the new double bridge tour will be used)
                self.shuffle = False

            else:

                # if double bridge was not found, next improvement loop will shuffle the tour
                self.shuffle = True

    def lk2_select_broken_edge(self, gain, t1, t2, t3, t4, broken_edges, joined_edges):
        """
        This function is used to select the next edge to be broken at current configuration of the LK optimization procedure. After that, the function will swap the tour nodes (t1,t2,t3,t4) and return a boolean value indicating if the swap resulted in a valid better tour.

        If the swap results in a valid worse tour, the call to select a new joined edge will be made to continue the exploration of new edges. This new exploration will try to swap the closed edge (t4,t1) with a new potencial better edge.

        The nodes t1, t2, t3 and t4 represents the swap in current configuration. When having sequence of swaps in the same tour, LK paper will mention those nodes based on the swap number, like (t2i, t2i+1, t2i+2...), where "i" is the swap number.

        :param gain: the current gain from last LK step
        :type gain: float
        :param t1: the tail node of the first broken edge in the 2-opt swap
        :type t1: Node2D
        :param t2: the head node of the first broken edge in the 2-opt swap
        :type t2: Node2D
        :param t3: the tail node of the second broken edge in the 2-opt swap
        :type t3: Node2D
        :param t4: the head node of the second broken edge in the 2-opt swap
        :type t4: Node2D
        :param broken_edges: the set of broken edges
        :type broken_edges: set
        :param joined_edges: the set of joined edges
        :type joined_edges: set
        :return: a boolean indicating if improvement was found
        :rtype: boolean
        """

        # get the edge to be broken and its cost
        broken_edge = Edge(t3, t4)
        broken_cost = self.cost_matrix[(t3.id, t4.id)]

        # the neighboor can't be t1 (this results in an invalid tour)
        # disjoint criteria (broken edge can't be previouslly at joined edges)
        # check if broken_edge is not already inside broken_edges set, because an edge can't be broken two times, this would lead to incorrect calculation
        if t1 != t4 and broken_edge not in joined_edges and broken_edge not in broken_edges:

            # check if swap is valid
            if (self.tour.is_swap_feasible(t1, t2, t3, t4)):

                    # execute the swap
                self.tour.swap_feasible(t1, t2, t3, t4)

                # add the broken edge to the set of broken edges
                broken_edges.add(broken_edge)

                # build the join edge from neighboor to last node (closing the tour)
                joined_edge = Edge(t4, t1)
                joined_cost = self.cost_matrix[(t4.id, t1.id)]

                # compute the current gain value
                curr_gain = gain + (broken_cost - joined_cost)

                # if gain is positive, restart the search with new tour
                # this is a simplification of LK from Helsgaun paper. LK suggest to continue the search, while Helsgaun suggest to simplify the algorithm
                if curr_gain > self.gain_precision:

                    # add joined edge
                    joined_edges.add(joined_edge)

                    # a sanity check here
                    old_tour_cost = self.tour.cost
                    self.tour.set_cost(self.cost_matrix)  # this shall not be erased in the future!
                    delta_tour_cost = abs(old_tour_cost - self.tour.cost)
                    delta_gain = abs(delta_tour_cost - curr_gain)

                    if (delta_gain > self.gain_precision):
                        self.logger.debug("delta gain error")

                    if (old_tour_cost < self.tour.cost):
                        self.logger.debug("cost error")

                    # return true value meaning an improvement was found
                    return True

                # if gain is negative, try to find a new best edge to add to form a new best tour
                else:

                    # try to find new joined edge
                    return self.lk2_select_joined_edge(curr_gain, t1, t4, broken_edges, joined_edges)

        # boolean value indicating no improvement was found
        return False

    def lk2_select_joined_edge(self, gain, t1, t4, broken_edges, joined_edges):
        """
        This function is used to select the next edge to be joined at current configuration of the LK optimization procedure. A false boolean is returned if a new joined edge was not found. If a replacement is found, the call for the selection of a new broken edge is made. 

        The selection function will "undo" the last joined edge (t4,t1) from previous "select_broken_edge" procedure, which closed the tour, by selecting a new potencial "node" so that (t1,t4) is broken and a new joined edge (t4,node) will be tested.

        :param gain: the current gain from last LK step
        :type gain: float
        :param t1: the head node of the last joined edge
        :type t1: Node2D
        :param t4: the tail node of the last joined edge
        :type t4: Node2D
        :param broken_edges: the set of broken edges
        :type broken_edges: set
        :param joined_edges: the set of joined edges
        :type joined_edges: set
        :return: a boolean indicating if improvement was found
        :rtype: boolean
        """

        # get the cost of the last joined edge (t4,t1)
        broken_cost = self.cost_matrix[(t4.id, t1.id)]

        # loop through the closest possible node from t4 instead of looping through all nodes
        for (node, neighbor_node), _ in self.get_best_neighbors(t4, t1):

            # create the edge and get the edge cost
            joined_edge = Edge(t4, node)
            joined_cost = self.cost_matrix[(t4.id, node.id)]

            # compute the new gain value
            curr_gain = gain + (broken_cost - joined_cost)

            # check if edge to be added is not in broken edges (disjoint criteria)
            # check if edge to be added is not already inside the tour
            # check if gain is positive
            if joined_edge not in broken_edges and joined_edge not in self.tour.edges and curr_gain > self.gain_precision:

                # add the edge to joined edges
                joined_edges.add(joined_edge)

                # try to select an edge to be broken
                return self.lk2_select_broken_edge(curr_gain, t1, t4, node, neighbor_node, broken_edges, joined_edges)

        # boolean value indicating no improvement was found
        return False

    def lk2_main(self):
        """
        Execute the main loop of LK Heuristic to optimize the tour. This is a single run of LK loop, which tries to find a better tour considering current tour being analyzed.

        :return: a boolean value indicating if an improve was found
        :rtype: boolean
        """

        # loop through each node in the tour, to apply the optimization
        # LK Heuristic will loop over all tour nodes as initial nodes (first node is explored in total)
        for t1 in self.tour.get_nodes():

            # loop through each neighboorhood node from initial node (second node is also explored as total, so at least 2-opt is always tested for all nodes)
            for t2 in (t1.pred, t1.succ):

                # get the break edge and the cost
                broken_edge = Edge(t1, t2)
                broken_cost = self.cost_matrix[(t1.id, t2.id)]

                # loop through the best possible nodes (t3,t4) from t2 instead of looping through all nodes
                for (t3, t4), _ in self.get_best_neighbors(t2, t1):

                    # get the joined edge and its cost
                    joined_edge = Edge(t3, t2)
                    joined_cost = self.cost_matrix[(t3.id, t2.id)]

                    # compute the gain
                    gain = broken_cost - joined_cost

                    # check if joined edge is not inside the existing tour
                    # check if gain is positive
                    if joined_edge not in self.tour.edges and gain > self.gain_precision:

                        # append the edges to each edges set
                        broken_edges = set([broken_edge])
                        joined_edges = set([joined_edge])

                        # try to select the next break edge and get a possibly new tour
                        if self.lk2_select_broken_edge(gain, t1, t2, t3, t4, broken_edges, joined_edges):

                            # reset the swap stack
                            self.tour.swap_stack.clear()

                            # return true value
                            return True
                        else:
                            # if new best tour was not found, restore the initial tour to continue the loop (if tour was not swapped, it will remain the same)
                            self.tour.restore()

        # if no improvement is found, return false
        return False

    def lk2_improve(self):
        """
        The improve loop using Lin-Kernighan Heuristic. The loop calls main LK optimizer everytime a better tour is found (which will try to optimize the new best tour using the same optimizer). The loop ends when no improvement is found, such that current tour will be the local optimal
        """

        # number of tours tested
        tour_count = 1

        # boolean checking if a improved tour was found
        improved = True

        # log the starting tour cost
        self.logger.debug(f"Starting tour cost: {self.tour.cost:.3f}")

        # loop until no improvement is found at TSP tour
        while improved:

            # compute the improvement
            improved = self.lk2_main()

            # log the current tour cost
            self.logger.debug(f"Current tour '{tour_count}' cost: {self.tour.cost:.3f}")

            # update tour count
            tour_count += 1

    def bf_improve(self):
        """
        The improve loop using Brute-Force computation, which converges for the global optimal tour. All permutations for nodes will be tested. It is recommended only for very small tours (< 10 nodes).
        """

        # assign values for the minimum cost and tour found
        min_cost = math.inf
        min_tour = None

        # create all possible permutations for the tour
        perms = permutations(self.nodes)

        # get the start node (to avoid repeated tours)
        start_node = self.nodes[0]

        # count number of tested tours
        tour_count = 0

        # loop through each permutation
        for perm in perms:

            # only permutation starting with the starting node is considered (so that repeated tours starting at a different node are not considered)
            if perm[0] == start_node:

                # loop through each node (starting from the last one)
                for i in range(-1, len(perm) - 1):

                    # get current and next nodes
                    curr_node = perm[i]
                    next_node = perm[i + 1]

                    # assign predecessor and successor nodes
                    curr_node.succ = next_node
                    next_node.pred = curr_node

                # update the cost value
                self.tour.set_cost(self.cost_matrix)

                # check if a best tour was found
                if self.tour.cost < min_cost:

                    # update best values
                    min_cost = self.tour.cost
                    min_tour = perm

                    # log the current  tour cost
                    self.logger.info(f"Current tour '{tour_count}'cost: {self.tour.cost:.3f}")

                # update and log count
                tour_count += 1
                if tour_count % 1000 == 0:
                    self.logger.info(f"Current tour '{tour_count}' cost: {min_cost:.3f}")

        # loop through each node of the best tour
        for i in range(-1, len(min_tour) - 1):

            # get current and next nodes
            curr_node = min_tour[i]
            next_node = min_tour[i + 1]

            # assign predecessor and successor nodes
            curr_node.succ = next_node
            next_node.pred = curr_node

        # update the cost value
        self.tour.set_cost(self.cost_matrix)

    def nn_improve(self):
        """
        The improve loop using Nearest-Neighbor computation.
        """

        # create a set of all tsp nodes
        nodes = set(self.nodes)

        # get the starting node
        curr_node = self.nodes[0]

        # start the set of visited nodes
        visited_nodes = set()
        visited_nodes.add(curr_node)

        # loop until all nodes are visited
        while (len(visited_nodes) < len(self.nodes)):

           # loop to find the nearest neighbor of current node
            cost = math.inf
            # loop through each node that were not visited yet
            for node in nodes - visited_nodes:
                if self.cost_matrix[(node.id, curr_node.id)] < cost:
                    cost = self.cost_matrix[(node.id, curr_node.id)]
                    next_node = node

            # update succ and pred for current and next nodes
            next_node.pred = curr_node
            curr_node.succ = next_node

            # add new node to visited nodes set
            visited_nodes.add(next_node)

            # update current node for next iteration
            curr_node = next_node

        # update the starting and ending nodes
        self.nodes[0].pred = curr_node
        curr_node.succ = self.nodes[0]

        # update the cost value
        self.tour.set_cost(self.cost_matrix)

    @classmethod
    def get_solution_methods(cls):
        """
        Get the list of solution methods for the tsp

        :return: list of solution method
        :rtype: list
        """
        return [cls.bf_improve, cls.nn_improve, cls.lk1_improve, cls.lk2_improve]



import os
import logging
import math
import time
from lk_heuristic.models.tsp import Tsp
from lk_heuristic.utils.cost_funcs import cost_funcs
from lk_heuristic.utils.io_funcs import import_tsp_file, export_tsp_file

# get the directory of this file and setup the "samples" and "solutions" dir
file_dir = os.path.dirname(os.path.realpath(__file__))
samples_dir = os.path.abspath(os.path.join(file_dir, "../", "samples"))
solutions_dir = os.path.abspath(os.path.join(file_dir, "../", "solutions"))


def get_interactive_inputs():
    """
    Execute the interactive shell with user inputs to get a selection for the tsp file (from 'samples' directory) and a selection of a solution method (a list of avaliable methods at Tsp class)

    :return: a tuple with the tsp file and the solution method selected by the user
    :rtype: tuple
    """

    ###################################
    # Step 1 - Select the TSP problem #
    ###################################

    # index of a file and collection of file paths
    tsp_files = []

    # loop to collect files in selected directory
    print("# Step 1 - TSP Instance Selection #")
    print(f"*TSP instances avaliable in '{os.path.basename(samples_dir)}' directory:")
    f_index = 0
    for f in os.listdir(samples_dir):
        if f.lower().endswith(".tsp"):
            tsp_files.append(os.path.join(samples_dir, f))
            print(f"[{f_index}] - {f}")
            f_index += 1

    # loop to get user selection
    tsp_selection = "-1"
    while (not tsp_selection.isnumeric() or int(tsp_selection) > len(tsp_files) - 1):
        tsp_selection = input("--> Select one of the instances: ")

    # get the tsp file
    tsp_file = tsp_files[int(tsp_selection)]

    # print separator
    print("")

    ###################################
    # Step 2 - Select solution method #
    ###################################

    # get the solution methods avaliable at Tsp class
    solution_methods = Tsp.get_solution_methods()

    # loop to display avaliable solution methods
    print("# Step 2 - TSP Solution Methods #")
    print("*TSP solution methods avaliable in tsp.py:")
    for i, solution_method in enumerate(solution_methods):
        print(f"[{i}] - {solution_method.__name__}")

    # loop to get user selection
    method_selection = "-1"
    while (not method_selection.isnumeric() or int(method_selection) > len(solution_methods) - 1):
        method_selection = input("--> Select one of the solution methods: ")

    # get the method name
    solution_method = solution_methods[int(method_selection)].__name__

    # print separator
    print("")

    # returns the tsp file and solution method
    return (tsp_file, solution_method)


def solve(tsp_file=None, solution_method=None, runs=1, backtracking=(5, 5), reduction_level=4, reduction_cycle=4, tour_type="cycle", file_name=None, logging_level=logging.DEBUG):
    """
    Solve a specific tsp problem a certain amount of times using the tsp_file input and the desired solution method. If this functions is called with no supplied inputs, the interactive inputs will be collected through the terminal. The best solution is parsed to .tsp file and exported to solution folder. 

    :param tsp_file: the .tsp file to be solved
    :type tsp_file: str
    :param solution_method: the method to be used when solving the tsp
    :type solution_method: function
    :param runs: the number of improve runs to be performed
    :type runs: int
    :param backtracking: the number of closest neighbors on each search level
    :type backtracking: tuple
    :param reduction_level: the search level where reduction starts being applied
    :type reduction_level: int
    :param reduction_cycle: the search cycle where reduction starts being applied
    :type reduction_cycle: int
    :param tour_type: the type of the tour (either 'path' or 'cycle')
    :type tour_type: str
    :param file_name: the name of the tsp file to be written
    :type file_name: str
    :param logging_level: the verbosity level for more or less details during execution
    :type logging_level: int
    """

    # get interactive inputs if input is not supplied at function
    if not (tsp_file or solution_method):
        tsp_file, solution_method = get_interactive_inputs()

    # setup the logger
    logging.basicConfig(level=logging_level, format='%(asctime)s [%(levelname)s] %(name)s: %(message)s')
    logger = logging.getLogger(__name__)

    # parse the .tsp file
    logger.info(f"Importing .tsp file '{os.path.basename(tsp_file)}'")
    tsp_header, tsp_nodes = import_tsp_file(tsp_file)

    # get the cost function
    cost_function = cost_funcs[tsp_header["EDGE_WEIGHT_TYPE"]]
    logger.info(f"Using '{cost_function.__name__}' for edge type '{tsp_header['EDGE_WEIGHT_TYPE']}'")

    # setup some information of the run
    best_tour = None  # the best tsp tour nodes found so far
    best_cost = math.inf  # the best cost found so fat
    mean_cost = 0  # the mean cost value of all runs

    # create the initial tsp instance
    logger.info("Creating TSP instance")
    tsp = Tsp(tsp_nodes, cost_function, shuffle=False, backtracking=backtracking, reduction_level=reduction_level, reduction_cycle=reduction_cycle, tour_type=tour_type, logging_level=logging_level)

    # looping through each run
    logger.info("Starting improve loop")
    for run in range(1, runs + 1):

        # shuffle the tour nodes
        if tsp.shuffle:
            tsp.tour.shuffle()

        # initialize tour cost
        tsp.tour.set_cost(tsp.cost_matrix)

        # execute the improvement method and timeit
        start_time = time.time()
        tsp.methods[solution_method]()
        end_time = time.time()

        # update best solution found so far (if current solution is the best)
        if tsp.tour.cost < best_cost:
            best_tour = tsp.tour.get_nodes()
            best_cost = tsp.tour.cost

        # update mean cost value
        mean_cost += (tsp.tour.cost - mean_cost) / run

        # log the information of current run
        logger.info(f"[Run:{run}] --> Cost: {tsp.tour.cost:.3f} / Best: {best_cost:.3f} / Mean: {mean_cost:.3f} ({end_time - start_time:.3f}s)")

    # setup export name as "tspname_cost.tsp"
    tsp_full_name = os.path.basename(tsp_file)
    tsp_name, tsp_ext = os.path.splitext(tsp_full_name)
    export_file = os.path.join(solutions_dir, f"{tsp_name}_{best_cost:.3f}.tsp")

    # if a file name is supplied, the export file is overwritten
    if file_name:
        export_file = os.path.join(solutions_dir, f"{file_name}.tsp")

    # execute the export
    logger.info(f"Exporting '{os.path.basename(export_file)}' file to solutions folder")
    export_tsp_file(export_file, tsp_header, best_tour, tour_type)



def euc_2d(n1, n2):
    """
    The euclidean distance for 2D cartesian nodes

    :param n1: first node
    :type n1: Node
    :param n2: second node
    :type n2: Node
    :return: the euclidean distance between first node and second node in 2D space
    :rtype: float
    """
    return ((n1.x - n2.x)**2 + (n1.y - n2.y)**2)**0.5


def euc_3d(n1, n2):
    """
    The euclidean distance for 3D cartesian nodes

    :param n1: first node
    :type n1: Node3D
    :param n2: second node
    :type n2: Node3D
    :return: the euclidean distance between first node and second node in 3D space
    :rtype: float
    """
    return ((n1.x - n2.x)**2 + (n1.y - n2.y)**2 + (n1.z - n2.z)**2)**0.5


# a dictionary mapping tsp lib edge weight type with respective cost function
cost_funcs = {"EUC_2D": euc_2d,
              "EUC_3D": euc_3d}



import unittest
from lk_heuristic.models.node import Node2D, Node3D


class TestNode(unittest.TestCase):
    """
    A class to perform testing on TSP nodes
    """

    def setUp(self):
        """
        The method to be performed before every test. Used to setup some nodes
        """

        # initialize for 2D nodes
        # a test node
        self.node_2d_1 = Node2D(0, 1)
        self.node_2d_1.id = 0

        # same node 1 creation
        self.node_2d_equal_1 = Node2D(0, 1)
        self.node_2d_equal_1.id = 0

        # a different node coordinates
        self.node_2d_2 = Node2D(2, 3)
        self.node_2d_2.id = 1

        # create same node coords, but change the id
        self.node_2d_almost_equal_1 = Node2D(0, 1)
        self.node_2d_almost_equal_1.id = 2

        # an empty node
        self.node_2d_empty = None

        # a reversed coordinate node
        self.node_2d_reversed_1 = Node2D(1, 0)
        self.node_2d_reversed_1.id = 3

        # initialize for 3D nodes
        # a test node
        self.node_3d_1 = Node3D(0, 1, 2)
        self.node_3d_1.id = 0

        # same node 1 creation
        self.node_3d_equal_1 = Node3D(0, 1, 2)
        self.node_3d_equal_1.id = 0

        # a different node coordinates
        self.node_3d_2 = Node3D(1, 1, 2)
        self.node_3d_2.id = 1

        # create same node coords, but change the id
        self.node_3d_almost_equal_1 = Node3D(0, 1, 2)
        self.node_3d_almost_equal_1.id = 2

        # an empty node
        self.node_3d_empty = None

        # a reversed coordinate node
        self.node_3d_reversed_1 = Node3D(2, 1, 0)
        self.node_3d_reversed_1.id = 3

        # almost similar nodes
        self.node_3d_same_x_y_1 = Node3D(0, 1, 3)
        self.node_3d_same_x_y_1.id = 4
        self.node_3d_same_y_z_1 = Node3D(3, 1, 2)
        self.node_3d_same_y_z_1.id = 5
        self.node_3d_same_x_z_1 = Node3D(0, 3, 2)
        self.node_3d_same_x_z_1.id = 6

    def test_node_2d_equality(self):
        """
        Testing node equality.
        """

        # similar node coords and ids, but different instances
        self.assertEqual(self.node_2d_1, self.node_2d_equal_1)

        # different node coords, but same id
        self.assertNotEqual(self.node_2d_1, self.node_2d_2)

        # same node coords, but different id
        self.assertNotEqual(self.node_2d_1, self.node_2d_almost_equal_1)

        # equality with an empty node
        self.assertNotEqual(self.node_2d_1, self.node_2d_empty)

        # reversed (x,y) coord values
        self.assertNotEqual(self.node_2d_1, self.node_2d_reversed_1)

    def test_node_2d_gt(self):
        """
        Testing node size comparison (greater than).
        """

        # different node coords
        self.assertLess(self.node_2d_1, self.node_2d_2)

        # reversed node coords
        self.assertLess(self.node_2d_1, self.node_2d_reversed_1)

    def test_node_3d_equality(self):
        """
        Testing node equality.
        """

        # similar node coords and ids, but different instances
        self.assertEqual(self.node_3d_1, self.node_3d_equal_1)

        # different node coords, but same id
        self.assertNotEqual(self.node_3d_1, self.node_3d_2)

        # same node coords, but different id
        self.assertNotEqual(self.node_3d_1, self.node_3d_almost_equal_1)

        # equality with an empty node
        self.assertNotEqual(self.node_3d_1, self.node_3d_empty)

        # reversed (x,y,z) coord values
        self.assertNotEqual(self.node_3d_1, self.node_3d_reversed_1)

        # 2 coords equal but one is different
        self.assertNotEqual(self.node_3d_1, self.node_3d_same_x_y_1)
        self.assertNotEqual(self.node_3d_1, self.node_3d_same_y_z_1)
        self.assertNotEqual(self.node_3d_1, self.node_3d_same_x_z_1)

    def test_node_3d_gt(self):
        """
        Testing node size comparison (greater than).
        """

        # different node coords
        self.assertLess(self.node_3d_1, self.node_3d_2)

        # reversed node coords
        self.assertLess(self.node_3d_1, self.node_3d_reversed_1)
